@@ -346,15 +346,27 @@ static int class_eq_event (lua_State* L)
 static int class_gc_event (lua_State* L)
 {
   int top = lua_gettop(L);
-	void* u = *((void**)lua_touserdata(L,1));
+	 void* u = *((void**)lua_touserdata(L,1));
   lua_pushstring(L,"tolua_gc");
-  lua_rawget(L,LUA_REGISTRYINDEX);
-  lua_pushlightuserdata(L,u);
-  lua_rawget(L,-2);
-  if (lua_isfunction(L,-1))
+  lua_rawget(L,LUA_REGISTRYINDEX);   /* gc */
+  lua_pushlightuserdata(L,u);        /* gc u */
+  lua_rawget(L,-2);                  /* gc func */
+  if (!lua_isnil(L, -1))
   {
-    lua_pushvalue(L,1);
-    lua_call(L,1,0);
+    /* remove entry from table */
+    lua_pushlightuserdata(L,u);
+    lua_pushnil(L);
+    lua_rawset(L,-4);
+    if (lua_isfunction(L,-1)) {
+      /* call collect function */
+      lua_pushvalue(L,1);            /* tolua_gc tolua_gc.u(func) u */
+      lua_call(L,1,0);               /* tolua_gc */
+    }
+    else if (lua_isuserdata(L,-1) && *((void**)lua_touserdata(L,-1))==NULL) {
+      /* free object */
+      free(u);
+      tolua_release(L,u);                /* unmap from tolua tables */
+    }
   }
 	lua_settop(L,top);
 	return 0;
