@@ -132,15 +132,10 @@ int tolua_bnd_takeownership (lua_State* L)
     {
       lua_pushstring(L,".collector");
       lua_rawget(L,-2);
-      func = lua_tocfunction(L,-1);
+      func = lua_iscfunction(L,-1) ? lua_tocfunction(L,-1) : NULL; 
       lua_pop(L,2);
-      if (func)
-      {
-        void* u = *((void**)lua_touserdata(L,1));
-        /* force garbage collection to avoid C to reuse a to-be-collected address */
-/*        lua_gc(L,LUA_GCCOLLECT,0); */
-        tolua_clone(L,u,func);
-      }
+      void* u = *((void**)lua_touserdata(L,1));
+      tolua_clone(L,u,func);
     }
   }
   lua_pushboolean(L,func!=0);
@@ -155,19 +150,14 @@ static int tolua_bnd_releaseownership (lua_State* L)
   if (lua_isuserdata(L,1))
   {
     void* u = *((void**)lua_touserdata(L,1));
-    /* force garbage collection to avoid releasing a to-be-collected address */
-    /*lua_gc(L,LUA_GCCOLLECT,0); */
     lua_pushstring(L,"tolua_gc");
     lua_rawget(L,LUA_REGISTRYINDEX);
     lua_pushlightuserdata(L,u);
     lua_rawget(L,-2);
-    if (lua_isfunction(L,-1))
-    {
-      lua_pushlightuserdata(L,u);
-      lua_pushnil(L);
-      lua_rawset(L,-4);
-      done = 1;
-    }
+    lua_pushlightuserdata(L,u);
+    lua_pushnil(L);
+    lua_rawset(L,-4);
+    done = 1;
   }
   lua_pushboolean(L,done!=0);
   return 1;
@@ -194,6 +184,18 @@ static int tolua_bnd_cast (lua_State* L)
   return 1;
 }
 
+/* Release
+** Function to be called by a Lua code that uses a function to explicitly
+** release a mapped object. This function is automatically called by all
+** destructors bound by tolua and by all collected objects.
+*/
+static int tolua_bnd_release (lua_State* L)
+{
+  void* value = tolua_tousertype(L,1,NULL);
+  if (value)
+    tolua_release(L,value);
+}
+
 TOLUA_API void tolua_open (lua_State* L)
 {
   int top = lua_gettop(L);
@@ -231,6 +233,7 @@ TOLUA_API void tolua_open (lua_State* L)
     tolua_function(L,"takeownership",tolua_bnd_takeownership);
     tolua_function(L,"releaseownership",tolua_bnd_releaseownership);
     tolua_function(L,"cast",tolua_bnd_cast);
+    tolua_function(L,"release",tolua_bnd_release);
     tolua_endmodule(L);
     tolua_endmodule(L);
   }
@@ -249,13 +252,20 @@ TOLUA_API void* tolua_copy (lua_State* L, void* value, unsigned int size)
   return clone;
 }
 
+
 /* Release userdata from tolua
  */
 TOLUA_API void tolua_release (lua_State* L, void* value)
 {
   lua_pushstring(L,"tolua_ubox");
-  lua_rawget(L,LUA_REGISTRYINDEX);        /* stack: ubox */
-  lua_pushlightuserdata(L,value);
+  lua_rawget(L,LUA_REGISTRYINDEX); /* stack: ubox */
+  lua_pushlightuserdata(L,value);  /* stack: ubox u */
+  /* set userdata pointer to NULL: this pointer might be reused by C/C++ */
+  lua_rawget(L,-2);                /* stack: ubox ud */
+  *((void**)lua_touserdata(L,1)) = NULL;
+  lua_pop(L,1);                    /* stack: ubox */
+  /* remove value from ubox */
+  lua_pushlightuserdata(L,value);  /* stack: ubox u */
   lua_pushnil(L);
   lua_rawset(L,-3);
   lua_pop(L,1);
